// vi: set ts=4 sw=4 :
// vim: set tw=75 :

// engine_api.cpp - implementation of Half-Life engine functions
/*
 * Copyright (c) 2004 OLO
 * This file is a part of PLBot application.
 * You are not allowed to remove these copyrights.
 */

#include <stdio.h>
#include <stdarg.h>

#include <extdll.h>

#include "engine_api.h"
#include "meta_api.h"
#include "plbot_api.h"
#include "bot.h"

extern int isFakeClientCommand;
extern int fake_arg_count;
extern enginefuncs_t g_engfuncs;
extern player_t bots[33];
extern float grenade_time;

void (*botMsgFunction)(void *, player_t*) = NULL;
void (*botMsgEndFunction)(void *, player_t*) = NULL;
player_t *botMsgIndex;

void Client_DeathMsg(void *p, player_t *pPlayer);
void Client_Damage(void *p, player_t *pPlayer);
void Client_CurWeapon(void *p, player_t *pPlayer);

int message_state = 0;

void MessageBegin(int msg_dest, int msg_type, const float *pOrigin, edict_t *ed)
{
   message_state = 0;   // reset message state
   botMsgFunction = NULL;     // no msg function until known otherwise
   botMsgEndFunction = NULL;  // no msg end function until known otherwise

   if (msg_type == GET_USER_MSG_ID(PLID, "DeathMsg", NULL))
      botMsgFunction = BotClient_DeathMsg;
   else if (msg_type == GET_USER_MSG_ID(PLID, "TextMsg", NULL))
      botMsgFunction = BotClient_TextMsg;
   else if (msg_type == GET_USER_MSG_ID(PLID, "WeaponList", NULL))
      botMsgFunction = BotClient_WeaponList;
   else if (msg_type == GET_USER_MSG_ID(PLID, "HLTV", NULL))
      botMsgFunction = BotClient_HLTV;
   else if (ed)
   {
      botMsgIndex = GET_EPLAYERPOINTER(ed);
      if (msg_type == GET_USER_MSG_ID(PLID, "VGUIMenu", NULL))
         botMsgFunction = BotClient_VGUIMenu;
      else if (msg_type == GET_USER_MSG_ID(PLID, "WeaponList", NULL))
         botMsgFunction = BotClient_WeaponList;
      else if (msg_type == GET_USER_MSG_ID(PLID, "CurWeapon", NULL))
         botMsgFunction = BotClient_CurrentWeapon;
      else if (msg_type == GET_USER_MSG_ID(PLID, "AmmoX", NULL))
         botMsgFunction = BotClient_AmmoX;
      else if (msg_type == GET_USER_MSG_ID(PLID, "AmmoPickup", NULL))
         botMsgFunction = BotClient_AmmoPickup;
      else if (msg_type == GET_USER_MSG_ID(PLID, "Damage", NULL))
         botMsgFunction = BotClient_Damage;
      else if (msg_type == GET_USER_MSG_ID(PLID, "Money", NULL))
         botMsgFunction = BotClient_Money;
      else if (msg_type == GET_USER_MSG_ID(PLID, "ScreenFade", NULL))
         botMsgFunction = BotClient_ScreenFade;
      else if (msg_type == GET_USER_MSG_ID(PLID, "SendAudio", NULL))
         botMsgFunction = BotClient_SendAudio;
      else if (msg_type == GET_USER_MSG_ID(PLID, "StatusIcon", NULL))
         botMsgFunction = BotClient_Icon;
      else if (msg_type == GET_USER_MSG_ID(PLID, "WeapPickup", NULL))
         botMsgFunction = BotClient_WeaponPickup;
      else if (msg_type == GET_USER_MSG_ID(PLID, "SetFOV", NULL))
         botMsgFunction = BotClient_SetFOV;
   }
   RETURN_META(MRES_IGNORED);
}

void MessageEnd(void)
{
   if (botMsgEndFunction)
      (*botMsgEndFunction)(NULL, botMsgIndex);  // NULL indicated msg end
   botMsgFunction = NULL;// clear out the bot message function pointers...
   botMsgEndFunction = NULL;
   RETURN_META(MRES_IGNORED);
}

void WriteByte(int iValue)
{
   if (botMsgFunction)
      (*botMsgFunction)((void *)&iValue, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}

void WriteChar(int iValue)
{
   if (botMsgFunction)
      (*botMsgFunction)((void *)&iValue, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}
void WriteShort(int iValue)
{
   if (botMsgFunction)
      (*botMsgFunction)((void *)&iValue, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}
void WriteLong(int iValue)
{
   if (botMsgFunction)
      (*botMsgFunction)((void *)&iValue, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}
void WriteAngle(float flValue)
{
   if (botMsgFunction)
      (*botMsgFunction)((void *)&flValue, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}
void WriteCoord(float flValue)
{
   if (botMsgFunction)
      (*botMsgFunction)((void *)&flValue, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}

bool EndRoundMessage(const char *msg)
{
   char messages[16][36] = {"#Terrorist_Escaped","#VIP_Not_Escaped","#Terrorists_Not_Escaped",
                            "#Hostages_Not_Rescued", "#Terrorists_Escaped","#Target_Saved","#All_Hostages_Rescued",
                            "#Round_Draw", "#Escaping_Terrorists_Neutralized","#Terrorists_Win","#CTs_Win",
                            "#Bomb_Defused", "#CTs_PreventEscape","#VIP_Escaped","#Target_Bombed","#VIP_Assassinated"};

   for (int index = 0; index < 16; index++)
      if (FStrEq(msg,messages[index]))
         return TRUE; //Text has been found

   return FALSE; //No there is no end roudn message
}

void WriteString(const char *sz)
{
   if (EndRoundMessage(sz))
   {
      //Go through bots and reset their arrays
      for (int index = 1; index < 33; index++)
         if (IsValidPlayer(&bots[index]))
            BotSpawnInit( &bots[index], RESPAWN_END_ROUND);

      t_group_number = 0;
      ct_group_number = 0;
      RoundGoing = FALSE; //Stop this round
      bombplanted = FALSE;
      bombdroped = FALSE;
      bombnotfound = FALSE;
   }
   if (botMsgFunction)
      (*botMsgFunction)((void *)sz, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}
void WriteEntity(int iValue)
{
   if (botMsgFunction)
      (*botMsgFunction)((void *)&iValue, botMsgIndex);
   RETURN_META(MRES_IGNORED);
}

void engClientCommand(edict_t *pEdict, char *szFmt, ...)
{
   RETURN_META(MRES_SUPERCEDE);
}

void SetClientMaxspeed(const edict_t *pEdict, float fNewMaxspeed)
{
   ((edict_t *)pEdict)->v.maxspeed = fNewMaxspeed;
   RETURN_META(MRES_IGNORED);
}

edict_t* FindEntityByString(edict_t *pEdictStartSearchAfter, const char *pszField, const char *pszValue)
{
   if (FStrEq(pszValue, "info_map_parameters"))
   {
      for (int i = 1; i <= gpGlobals->maxClients; i++)
      {
         player_t *pBot = GET_IPLAYERPOINTER(i);
         if (!IsValidPlayer(pBot) || !(pBot->pEdict->v.flags & FL_FAKECLIENT))
            continue;
         pBot->buy_state = 0;
         BotSpawnInit( pBot, RESPAWN_START_ROUND);
      }
      RoundGoing = TRUE; //Start this round
   }

   RETURN_META_VALUE(MRES_IGNORED, 0);
}

const char *Cmd_Args( void )
{
   if (isFakeClientCommand)
      RETURN_META_VALUE(MRES_SUPERCEDE, &g_argv[0]);
   else
      RETURN_META_VALUE(MRES_IGNORED, NULL);
}

const char *Cmd_Argv( int argc )
{
   if (isFakeClientCommand)
   {
      switch (argc)
      {
      case 0:
         RETURN_META_VALUE(MRES_SUPERCEDE, &g_argv[64]);
         break;
      case 1:
         RETURN_META_VALUE(MRES_SUPERCEDE, &g_argv[128]);
         break;
      case 2:
         RETURN_META_VALUE(MRES_SUPERCEDE, &g_argv[192]);
         break;
      default://Original HL always return pointer on empty (not null) space
         RETURN_META_VALUE(MRES_SUPERCEDE, "");
      }
   }
   else
      RETURN_META_VALUE(MRES_IGNORED, NULL);
}


int Cmd_Argc( void )
{
   if (isFakeClientCommand)
      RETURN_META_VALUE(MRES_SUPERCEDE, fake_arg_count);
   else
      RETURN_META_VALUE(MRES_IGNORED, 0);
}

enginefuncs_t meta_engfuncs = {
                                 NULL,   // pfnPrecacheModel()
                                 NULL,   // pfnPrecacheSound()
                                 NULL,    // pfnSetModel()
                                 NULL,    // pfnModelIndex()
                                 NULL,   // pfnModelFrames()

                                 NULL,    // pfnSetSize()
                                 NULL,   // pfnChangeLevel()
                                 NULL,   // pfnGetSpawnParms()
                                 NULL,   // pfnSaveSpawnParms()

                                 NULL,    // pfnVecToYaw()
                                 NULL,   // pfnVecToAngles()
                                 NULL,   // pfnMoveToOrigin()
                                 NULL,    // pfnChangeYaw()
                                 NULL,   // pfnChangePitch()

                                 FindEntityByString,  // pfnFindEntityByString()
                                 NULL,   // pfnGetEntityIllum()
                                 NULL,  // pfnFindEntityInSphere()
                                 NULL,  // pfnFindClientInPVS()
                                 NULL,   // pfnEntitiesInPVS()

                                 NULL,   // pfnMakeVectors()
                                 NULL,   // pfnAngleVectors()

                                 NULL,   // pfnCreateEntity()
                                 NULL,   // pfnRemoveEntity()
                                 NULL,  // pfnCreateNamedEntity()

                                 NULL,    // pfnMakeStatic()
                                 NULL,   // pfnEntIsOnFloor()
                                 NULL,   // pfnDropToFloor()

                                 NULL,    // pfnWalkMove()
                                 NULL,    // pfnSetOrigin()

                                 NULL,    // pfnEmitSound()
                                 NULL,  // pfnEmitAmbientSound()

                                 NULL,    // pfnTraceLine()
                                 NULL,    // pfnTraceToss()
                                 NULL,  // pfnTraceMonsterHull()
                                 NULL,    // pfnTraceHull()
                                 NULL,    // pfnTraceModel()
                                 NULL,   // pfnTraceTexture()
                                 NULL,   // pfnTraceSphere()
                                 NULL,   // pfnGetAimVector()

                                 NULL,   // pfnServerCommand()
                                 NULL,   // pfnServerExecute()
                                 engClientCommand,  // pfnClientCommand() // d'oh, ClientCommand in dllapi too (eng)

                                 NULL,   // pfnParticleEffect()
                                 NULL,    // pfnLightStyle()
                                 NULL,    // pfnDecalIndex()
                                 NULL,   // pfnPointContents()

                                 MessageBegin,   // pfnMessageBegin()
                                 MessageEnd,    // pfnMessageEnd()

                                 WriteByte,    // pfnWriteByte()
                                 WriteChar,    // pfnWriteChar()
                                 WriteShort,    // pfnWriteShort()
                                 WriteLong,    // pfnWriteLong()
                                 WriteAngle,    // pfnWriteAngle()
                                 WriteCoord,    // pfnWriteCoord()
                                 WriteString,   // pfnWriteString()
                                 WriteEntity,   // pfnWriteEntity()

                                 NULL,   // pfnCVarRegister()
                                 NULL,   // pfnCVarGetFloat()
                                 NULL,   // pfnCVarGetString()
                                 NULL,   // pfnCVarSetFloat()
                                 NULL,   // pfnCVarSetString()

                                 NULL,   // pfnAlertMessage()
                                 NULL,   // pfnEngineFprintf()

                                 NULL, // pfnPvAllocEntPrivateData()
                                 NULL,  // pfnPvEntPrivateData()
                                 NULL,  // pfnFreeEntPrivateData()

                                 NULL,   // pfnSzFromIndex()
                                 NULL,   // pfnAllocString()

                                 NULL,    // pfnGetVarsOfEnt()
                                 NULL,  // pfnPEntityOfEntOffset()
                                 NULL,  // pfnEntOffsetOfPEntity()
                                 NULL,   // pfnIndexOfEdict()
                                 NULL,  // pfnPEntityOfEntIndex()
                                 NULL,  // pfnFindEntityByVars()
                                 NULL,   // pfnGetModelPtr()

                                 NULL,    // pfnRegUserMsg()

                                 NULL,  // pfnAnimationAutomove()
                                 NULL,  // pfnGetBonePosition()

                                 NULL,  // pfnFunctionFromName()
                                 NULL,  // pfnNameForFunction()

                                 NULL,   // pfnClientPrintf()   //! JOHN: engine callbacks so game DLL can print messages to individual clients
                                 NULL,   // pfnServerPrint()

                                 Cmd_Args,    // pfnCmd_Args()    //! these 3 added
                                 Cmd_Argv,    // pfnCmd_Argv()    //! so game DLL can easily
                                 Cmd_Argc,    // pfnCmd_Argc()    //! access client 'cmd' strings

                                 NULL,   // pfnGetAttachment()

                                 NULL,    // pfnCRC32_Init()
                                 NULL, // pfnCRC32_ProcessBuffer()
                                 NULL,  // pfnCRC32_ProcessByte()
                                 NULL,   // pfnCRC32_Final()

                                 NULL,    // pfnRandomLong()
                                 NULL,   // pfnRandomFloat()

                                 NULL,    // pfnSetView()
                                 NULL,     // pfnTime()
                                 NULL,   // pfnCrosshairAngle()

                                 NULL,   // pfnLoadFileForMe()
                                 NULL,    // pfnFreeFile()

                                 NULL,    // pfnEndSection()    //! trigger_endsection
                                 NULL,  // pfnCompareFileTime()
                                 NULL,    // pfnGetGameDir()
                                 NULL, // pfnCvar_RegisterVariable()
                                 NULL,  // pfnFadeClientVolume()
                                 SetClientMaxspeed,  // pfnSetClientMaxspeed()
                                 NULL,  // pfnCreateFakeClient()   //! returns NULL if fake client can't be created
                                 NULL,   // pfnRunPlayerMove()
                                 NULL,  // pfnNumberOfEntities()

                                 NULL,  // pfnGetInfoKeyBuffer()  //! passing in NULL gets the serverinfo
                                 NULL,   // pfnInfoKeyValue()
                                 NULL,   // pfnSetKeyValue()
                                 NULL,  // pfnSetClientKeyValue()

                                 NULL,    // pfnIsMapValid()
                                 NULL,   // pfnStaticDecal()
                                 NULL,  // pfnPrecacheGeneric()
                                 NULL,   // pfnGetPlayerUserId()   //! returns the server assigned userid for this player.
                                 NULL,   // pfnBuildSoundMsg()
                                 NULL,  // pfnIsDedicatedServer()  //! is this a dedicated server?
                                 NULL,   // pfnCVarGetPointer()
                                 NULL,   // pfnGetPlayerWONId()   //! returns the server assigned WONid for this player.

                                 //! YWB 8/1/99 TFF Physics additions
                                 NULL,   // pfnInfo_RemoveKey()
                                 NULL,  // pfnGetPhysicsKeyValue()
                                 NULL,  // pfnSetPhysicsKeyValue()
                                 NULL, // pfnGetPhysicsInfoString()
                                 NULL,   // pfnPrecacheEvent()
                                 NULL,   // pfnPlaybackEvent()

                                 NULL,    // pfnSetFatPVS()
                                 NULL,    // pfnSetFatPAS()

                                 NULL,  // pfnCheckVisibility()

                                 NULL,   // pfnDeltaSetField()
                                 NULL,  // pfnDeltaUnsetField()
                                 NULL,  // pfnDeltaAddEncoder()
                                 NULL,  // pfnGetCurrentPlayer()
                                 NULL,   // pfnCanSkipPlayer()
                                 NULL,   // pfnDeltaFindField()
                                 NULL, // pfnDeltaSetFieldByIndex()
                                 NULL, // pfnDeltaUnsetFieldByIndex()

                                 NULL,   // pfnSetGroupMask()

                                 NULL, // pfnCreateInstancedBaseline()  // d'oh, CreateInstancedBaseline in dllapi too
                                 NULL,   // pfnCvar_DirectSet()

                                 NULL,  // pfnForceUnmodified()

                                 NULL,   // pfnGetPlayerStats()

                                 NULL,  // pfnAddServerCommand()

                                 // Added in SDK 2.2:
                                 NULL, // pfnVoice_GetClientListening()
                                 NULL, // pfnVoice_SetClientListening()
                              };

C_DLLEXPORT int GetEngineFunctions(enginefuncs_t *pengfuncsFromEngine, int *interfaceVersion )
{
   LOG_DEVELOPER(PLID, "called: GetEngineFunctions; version=%d", *interfaceVersion);
   if(!pengfuncsFromEngine)
   {
      LOG_ERROR(PLID, "GetEngineFunctions called with null pengfuncsFromEngine");
      return(FALSE);
   }
   else if(*interfaceVersion != ENGINE_INTERFACE_VERSION)
   {
      LOG_ERROR(PLID, "GetEngineFunctions version mismatch; requested=%d ours=%d", *interfaceVersion, ENGINE_INTERFACE_VERSION);
      // Tell metamod what version we had, so it can figure out who is out of date.
      *interfaceVersion = ENGINE_INTERFACE_VERSION;
      return(FALSE);
   }
   memcpy(pengfuncsFromEngine, &meta_engfuncs, sizeof(enginefuncs_t));
   return TRUE;
}
