// vi: set ts=4 sw=4 :
// vim: set tw=75 :

// sdk_util.cpp - utility routines from HL SDK util.cpp

// Selected portions of dlls/util.cpp from SDK 2.1.
// Functions copied from there as needed...

/***
*
* Copyright (c) 1999, 2000 Valve LLC. All rights reserved.
*
* This product contains software technology licensed from Id
* Software, Inc. ("Id Technology").  Id Technology (c) 1996 Id Software, Inc.
* All Rights Reserved.
*
*   Use, distribution, and modification of this source code and/or resulting
*   object code is restricted to non-commercial enhancements to products from
*   Valve LLC.  All other use, distribution, or modification is prohibited
*   without written permission from Valve LLC.
*
****/
/*
 
===== util.cpp ========================================================
 
  Utility code.  Really not optional after all.
 
*/

/*
Last modify: 28/12/2001 21:50
By: OLO
*/

/*
 * Copyright (c) 2004 OLO
 * This file is a part of PLBot application.
 * You are not allowed to remove these copyrights.
 */

#include <extdll.h>
#include "sdk_util.h"
#include <cbase.h>
#include <string.h>   // for strncpy(), etc
#include "plbot_api.h" // me
#include "ctype.h"

#include <dllapi.h>
#include <meta_api.h>

char* UTIL_VarArgs( char *format, ... )
{
   va_list  argptr;
   static char  string[1024];
   va_start (argptr, format);
   vsprintf (string, format,argptr);
   va_end (argptr);
   return string;
}


//=========================================================
// UTIL_LogPrintf - Prints a logged message to console.
// Preceded by LOG: ( timestamp ) < message >
//=========================================================
void UTIL_LogPrintf( char *fmt, ... )
{
   va_list   argptr;
   static char  string[1024];
   va_start ( argptr, fmt );
   vsprintf ( string, fmt, argptr );
   va_end   ( argptr );
   // Print to server console
   ALERT( at_logged, "%s", string );
}



int gmsgTextMsg = 0;
int gmsgSayText = 0;
int gmsgShowMenu = 0;

//PLBOT menu
const char *BotMenu =
   {"\\yPLBot Menu\\w\n\n1.  Add Bot to T Team\n2.  Add Bot to CT Team\n3.  Set Bot Skill\n\n\\%s4.  Slap Bot Menu\n5.  Kick Bot Menu\n\n\\w6.  Nodes Editor\n\\%s7.  Game Settings\n\\%s8.  Custom Commands\n\n\\w0.  Exit"
   };

const char *BotNodeMenu =
   {"\\yNodes Editor\\w\n\n1.  Add Node\n2.  Remove Node\n3.  Set Rush Point\n\n4.  Save Nodes\n\n5.  %s Nodes\n\n0.  Back"
   };

const char *BotSkillMenu =
   {"\\yBot Skill Menu\\w\n\n\\w1.  Start from begin (easy)\\R\\y%s\n\\w2.  I can win\\R\\y%s\n\\w3.  Play with me\\R\\y%s\n\\w4.  Hurt me plenty\\R\\y%s\n\\w5.  Terrorist hunting (hard)\\R\\y%s\n\n\\w6.  Randomized skill\\R\\y%s\n\n\\w0.  Back"
   };

Vector UTIL_VecToAngles( const Vector &vec )
{
   float rgflVecOut[3];
   VEC_TO_ANGLES(vec, rgflVecOut);
   return Vector(rgflVecOut);
}

// Overloaded to add IGNORE_GLASS
void UTIL_TraceLine( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, IGNORE_GLASS ignoreGlass, edict_t *pentIgnore, TraceResult *ptr )
{
   TRACE_LINE( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE) | (ignoreGlass?0x100:0), pentIgnore, ptr );
}

void UTIL_TraceLine( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, edict_t *pentIgnore, TraceResult *ptr )
{
   TRACE_LINE( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE), pentIgnore, ptr );
}

void UTIL_TraceHull( const Vector &vecStart, const Vector &vecEnd, IGNORE_MONSTERS igmon, int hullNumber, edict_t *pentIgnore, TraceResult *ptr )
{
   TRACE_HULL( vecStart, vecEnd, (igmon == ignore_monsters ? TRUE : FALSE), hullNumber, pentIgnore, ptr );
}

edict_t *UTIL_FindEntityInSphere( edict_t *pentStart, const Vector &vecCenter, float flRadius )
{
   edict_t  *pentEntity;

   pentEntity = FIND_ENTITY_IN_SPHERE( pentStart, vecCenter, flRadius);

   if (!FNullEnt(pentEntity))
      return pentEntity;

   return NULL;
}

edict_t *UTIL_FindEntityByString( edict_t *pentStart, const char *szKeyword, const char *szValue )
{
   edict_t *pentEntity;

   pentEntity = FIND_ENTITY_BY_STRING( pentStart, szKeyword, szValue );

   if (!FNullEnt(pentEntity))
      return pentEntity;
   return NULL;
}

edict_t *UTIL_FindEntityByClassname( edict_t *pentStart, const char *szName )
{
   return UTIL_FindEntityByString( pentStart, "classname", szName );
}

edict_t *UTIL_FindEntityByModelname( edict_t *pentStart, const char *szName )
{
   return UTIL_FindEntityByString( pentStart, "model", szName );
}

#ifdef   DEBUG
edict_t *DBG_EntOfVars( const entvars_t *pev )
{
   if (pev->pContainingEntity != NULL)
      return pev->pContainingEntity;
   ALERT(at_console, "entvars_t pContainingEntity is NULL, calling into engine");
   edict_t* pent = (*g_engfuncs.pfnFindEntityByVars)((entvars_t*)pev);
   if (pent == NULL)
      ALERT(at_console, "DAMN!  Even the engine couldn't FindEntityByVars!");
   ((entvars_t *)pev)->pContainingEntity = pent;
   return pent;
}
#endif //DEBUG

bool FInViewCone(Vector *pOrigin, edict_t *pEdict, float deg)
{
   Vector2D vec2LOS;
   float    flDot;

   MAKE_VECTORS ( pEdict->v.angles );

   vec2LOS = ( *pOrigin - pEdict->v.origin ).Make2D();
   vec2LOS = vec2LOS.Normalize();

   flDot = DotProduct (vec2LOS , gpGlobals->v_forward.Make2D() );

   if ( flDot > deg )  // 0.50 60 degree field of view
      return TRUE;
   else
      return FALSE;
}

bool FInViewConeOfRadar(Vector *pOrigin, edict_t *pEdict, float deg)
{
   Vector2D vec2LOS;
   float    flDot;

   //UTIL_MakeVectors ( pEdict->v.v_angle );
   MAKE_VECTORS ( pEdict->v.movedir );

   vec2LOS = ( *pOrigin - pEdict->v.origin ).Make2D();
   vec2LOS = vec2LOS.Normalize();

   flDot = DotProduct (vec2LOS , gpGlobals->v_forward.Make2D() );

   if ( flDot > deg )  // 0.50 60 degree field of view
      return TRUE;
   else
      return FALSE;
}

bool FVisible( const Vector &vecOrigin, edict_t *pEdict )
{
   TraceResult tr;
   Vector      vecLookerOrigin;

   // look through caller's eyes
   vecLookerOrigin = pEdict->v.origin + pEdict->v.view_ofs;

   int bInWater = (POINT_CONTENTS (vecOrigin) == CONTENTS_WATER);
   int bLookerInWater = (POINT_CONTENTS (vecLookerOrigin) == CONTENTS_WATER);

   // don't look through water
   if (bInWater != bLookerInWater)
      return FALSE;

   UTIL_TraceLine(vecLookerOrigin, vecOrigin, ignore_monsters, ignore_glass, pEdict, &tr);

   if (tr.flFraction != 1.0)
      return FALSE;  // Line of sight is not established
   else
      return TRUE;  // line of sight is valid.
}

#define MAX_MENU_CHUNK 176
extern int menu_keys[10];

void UTIL_ShowMenu( edict_t *pEdict, int slots, int displaytime, bool needmore, char *aMenuList )
{
   char *pMenuList = aMenuList;
   char chunk[MAX_MENU_CHUNK+1];
   int char_count = 0;

   //Here you must set keys which will be available for player
   player_t *pPlayer = GET_EPLAYERPOINTER(pEdict);
   pPlayer->menu_keys = 0; //Reset menu keys bits

   if (gmsgShowMenu == 0)
      gmsgShowMenu = REG_USER_MSG( "ShowMenu", -1 );

   for (int i = 0; i < 10;i++)
   {//Go thorugh the table and set key bits
      if (slots & menu_keys[i])
         pPlayer->menu_keys |= (1<<(i+1));
   }

   while ( pMenuList && *pMenuList )
   {
      if ( strlen( pMenuList ) < MAX_MENU_CHUNK )
      {
         strcpy( chunk, pMenuList );
      }
      else
      {
         strncpy( chunk, pMenuList, MAX_MENU_CHUNK );
         chunk[MAX_MENU_CHUNK] = 0;  // strncpy doesn't always append the null terminator
      }

      //Move text to next chunk
      char_count += strlen( chunk );
      pMenuList = aMenuList + char_count;

      MESSAGE_BEGIN( MSG_ONE, gmsgShowMenu, NULL, pEdict );
      WRITE_SHORT( slots );
      WRITE_CHAR( displaytime );
      WRITE_BYTE(*pMenuList ? TRUE : FALSE); //Need more menu?
      WRITE_STRING(chunk);
      MESSAGE_END();
   }
}

//Get player names for players list
void GetPlayersList(player_t *pPlayer)
{
   edict_t *pClient = NULL;
   pPlayer->players_num = 0; //Reset number of names (palyers)

   for (int a = 1; a <=gpGlobals->maxClients; a++)
   {
      pClient = INDEXENT(a);
      if (!IsValidPlayer(pClient) || pClient == pPlayer->pEdict)
         continue;
      if (pClient->v.flags & FL_FAKECLIENT)
      {
         pPlayer->pList[pPlayer->players_num] = pClient;
         pPlayer->players_num++;
      }
   }
}

void MenuPlayersList(player_t *pPlayer, const char *menu_header, int pnum, int itype, int state)
{
   GetPlayersList(pPlayer); //Check if there is someone to kick

   if (pPlayer->players_num <= 0)
   {
      Botmenu(pPlayer, MSG_CS_BOT_MENU);
      return;
   }

   int list_end, pages_num = 0, list_begin, index_key = 0, keys = 0;
   char tmp[256],textmenu[512], listname[128];
   edict_t *pSelect;

   while (pages_num*pnum < pPlayer->players_num)
      pages_num++;

   sprintf(textmenu, "\\y%s %d/%d\\w\n\n", menu_header, state + 1, pages_num); //Print header

   //Set state and get list begin
   pPlayer->menu_state = state;
   list_begin = state*pnum;

   //Does our list have to now or later...
   if (pPlayer->players_num > list_begin + pnum)
      list_end = list_begin + pnum;
   else
      list_end = pPlayer->players_num;

   for (int i = list_begin; i < list_end; i++)
   {
      pSelect = pPlayer->pList[i];
      strcpy(listname,STRING(pSelect->v.netname));
      //Player not exist, is dead or is admin (two last cases are swithc able)
      if (!IsAlive(pSelect) && itype == MSG_CS_SLAP_MENU)
      {
         sprintf(tmp, "\\d%d.  %s\\w\n", index_key+1, listname);
         strcat(textmenu, tmp);
      }
      else
      {
         if  ((itype == MSG_CS_SLAP_MENU) && (pSelect->v.health <= pPlayer->slap_power))
            sprintf(tmp, "\\w%d.  %s  \\R\\ykill         \n", index_key+1, listname);
         else
            sprintf(tmp, "\\w%d.  %s\n", index_key+1, listname);

         strcat(textmenu, tmp);
         keys += menu_keys[index_key]; //Add proper key to menu keys
      }

      index_key++;
   }

   if (itype == MSG_CS_SLAP_MENU)
   {
      keys += menu_keys[6];
      sprintf(tmp, "\n\\w7.  Slap all bots");
      strcat(textmenu, tmp); //End msg

      keys += menu_keys[7];
      sprintf(tmp, "\n\\w8.  Slap damage: %d\n", pPlayer->slap_power);
      strcat(textmenu, tmp); //End msg
   }

   if (list_end < pPlayer->players_num)
   {
      strcat(textmenu,"\n\\w9.  More...");
      keys += menu_keys[8];
   }

   strcat(textmenu,"\n\\w0.  Back\n");
   keys += menu_keys[9];

   UTIL_ShowMenu(pPlayer->pEdict, keys, -1, FALSE,textmenu);
   pPlayer->start_action = itype;
}

//Show execute menu
void MenuCmdList(player_t *pPlayer, int state)
{
   int list_end, pages_num = 0, list_begin, index_key = 0, keys = menu_keys[9]; //reset keys
   char textmenu[512], tmp[256];

   while (pages_num*8 <  cstcmd_num)
      pages_num++;

   sprintf(textmenu, "\\yCustom Commands %d/%d\\w\n\n",state + 1, pages_num);

   pPlayer->menu_state = state;
   list_begin = state*8;

   if (cstcmd_num > list_begin + 8)
      list_end = list_begin + 8;
   else
      list_end = cstcmd_num;

   for (int i = list_begin; i < list_end; i++)
   {

      //If there is only one space as command make it as blank line
      if ((cstcmd_names[i].cmd[1] == 0) && (cstcmd_names[i].cmd[0] == ' '))
         sprintf(tmp, "\\d    %s\\w\n", cstcmd_names[i].cmd_d);
      else
      {
         sprintf(tmp, "%d.  %s\n", index_key+1, cstcmd_names[i].cmd_d);
         keys += menu_keys[index_key];
      }
      strcat(textmenu, tmp);  //Add proper key to menu keys
      index_key++;
   }

   if (list_end < cstcmd_num)
   {
      keys += menu_keys[8];
      strcat(textmenu,"\n\\w9.  More...");
   }

   strcat(textmenu,"\n\\w0.  Back\n");
   pPlayer->start_action = MSG_CS_CMD_MENU;
   UTIL_ShowMenu(pPlayer->pEdict, keys, -1,FALSE, textmenu );
}

//Settings Menu
void MenuSettingsList(player_t *pPlayer, int state)
{
   int list_end, list_begin, index_key = 0;
   int keys = menu_keys[9]; //reset keys
   char tmp[256], textmenu[512];
   int pages_num = 0;
   const char *break_menu =
      {"                  "
      };

   while (pages_num*8 < cvars_num)
      pages_num++;

   sprintf(textmenu, "\\yGame Settings %d/%d\\w\n\n", state + 1, pages_num);

   pPlayer->menu_state = state;
   list_begin = state*8;

   if (cvars_num > list_begin + 8)
      list_end = list_begin + 8;
   else
      list_end = cvars_num;

   for (int i = list_begin; i < list_end; i++)
   {
      float set_value = CVAR_GET_FLOAT(cvar_game[i].name);

      if ((set_value - (int)set_value) == 0)
         sprintf(tmp, "\\w%d.  %s\\R%5.0f%s\n", index_key+1, cvar_game[i].name, set_value,break_menu);
      else
         sprintf(tmp, "\\w%d.  %s\\R%5.2f%s\n", index_key+1, cvar_game[i].name, set_value,break_menu);

      keys += menu_keys[index_key]; //Add proper key to menu keys
      strcat(textmenu, tmp);
      index_key++;
   }

   if (list_end < cvars_num)
   {
      keys += menu_keys[8];
      strcat(textmenu,"\n\\w9.  More...");
   }

   strcat(textmenu,"\n\\w0.  Back");

   UTIL_ShowMenu(pPlayer->pEdict, keys, -1,FALSE,textmenu );
   pPlayer->start_action = MSG_CS_SET_MENU;
}

bool BotsExists()
{
   player_t *pPlayer =NULL;
   for (int a = 1; a <= gpGlobals->maxClients;a++)
   {
      pPlayer = GET_IPLAYERPOINTER(a);
      if (IsValidPlayer(pPlayer) && (pPlayer->pEdict->v.flags & FL_FAKECLIENT))
         return TRUE;
   }

   return FALSE;
}

void Botmenu(player_t *pPlayer, int imenu)
{
   pPlayer->start_action = imenu; //Set menu flag

   //Execute menu
   switch(imenu)
   {//Check the cases
   case MSG_CS_BOT_MENU: //Standard bot menu
      {int keys = 767;
         char text_menu[256];
         char cstcmd_tag[4] = "w";
         char cvars_tag[4] = "w";
         char bot_tag[4] = "w";
         if (!BotsExists())
         {
            strcpy(bot_tag, "d");
            keys -= 24;
         }
         if (cvars_num <= 0)
         {
            keys -= 64;
            strcpy(cvars_tag, "d");
         }
         if (cstcmd_num <= 0)
         {
            keys -= 128;
            strcpy(cstcmd_tag, "d");
         }
         sprintf(text_menu,BotMenu,bot_tag, cvars_tag,cstcmd_tag);
         UTIL_ShowMenu(pPlayer->pEdict, keys, -1, FALSE,text_menu ); //Addon Menu
         break;
      }
   case MSG_CS_NODE_MENU: //Nodes editor menu
      { char text_menu[192];
         char text_m[8] = "Hide";
         if (nodes_on)
            strcpy(text_m ,"Hide");
         else
            strcpy(text_m ,"Show");
         sprintf(text_menu,BotNodeMenu, text_m);
         UTIL_ShowMenu(pPlayer->pEdict, 543, -1, FALSE, text_menu); //Addon Menu
         break;
      }
   case MSG_CS_SKILL_MENU: //Skill menu
      { char text_menu[192];
         char skill1_tag[16] = " ";
         char skill2_tag[16] = " ";
         char skill3_tag[16] = " ";
         char skill4_tag[16] = " ";
         char skill5_tag[16] = " ";
         char skill6_tag[16] = " ";
         int skill = (int)plb_skill->value;
         if (skill == 1)
            strcpy(skill1_tag, "(((         ");
         if (skill == 2)
            strcpy(skill2_tag, "(((         ");
         if (skill == 3)
            strcpy(skill3_tag, "(((         ");
         if (skill == 4)
            strcpy(skill4_tag, "(((         ");
         if (skill == 5)
            strcpy(skill5_tag, "(((         ");
         if (skill == 6)
            strcpy(skill6_tag, "(((         ");
         sprintf(text_menu,BotSkillMenu,skill1_tag,skill2_tag,skill3_tag,skill4_tag,skill5_tag,skill6_tag );
         UTIL_ShowMenu(pPlayer->pEdict, 575, -1, FALSE, text_menu ); //Addon Menu
         break;
      }
   }
}
